#include "DataProcessorFactory.hpp"
#include "DataProcessor32.hpp"
#include "DataProcessor64.hpp"
#include <scTDC_types.h>


DataProcessorFactoryException::DataProcessorFactoryException
(const char *msg, int code)
: msg_(msg), code_(code)
{}

DataProcessorFactoryException::~DataProcessorFactoryException() throw()
{}

const char *
DataProcessorFactoryException::what() const throw()
{
	return msg_.c_str();
}

int
DataProcessorFactoryException::code() const
{
	return code_;
}



iDataProcessor *DataProcessorFactory::create(const sc_tdc_format &format, HistoData &h)
{
	/*
	 * create appripriate data_processor. settings are from format
	 * does not support ref counting. return values must be deleted explicitly with
	 * delete op
	 */
	iDataProcessor *ret;
	switch (format.total_bits_length) {
		case 8:
			throw DataProcessorFactoryException("bit length not implemented", -3);
			//no break;
		case 16:
			throw DataProcessorFactoryException("bit length not implemented", -2);
			//no break;
		case 32:
			ret = new DataProcessor32(h, format);
			break;
		case 64:
			ret = new DataProcessor64(h, format);
			break;
		default:
			throw DataProcessorFactoryException("unsupported bit length", -1);
			//no break;
	}

	return ret;
}
