#include "MainWindow.hpp"
#include <QStatusBar>
#include <QSettings>
#include <QApplication>
#include "AboutDialog.hpp"

MainWindow::MainWindow(QWidget *p)
: QMainWindow(p)
{
	setWindowTitle("TdcDemo");
	menu_bar = menuBar();
	menu_file = new QMenu("File", this);
	menu_device = new QMenu("Device", this);
	menu_help = new QMenu("Help", this);
	data_export = new QAction("Export", this);
	menu_file->addAction(data_export);
	quit = new QAction("Quit", this);
	menu_file->addAction(quit);
	about = new QAction("About", this);
	menu_help->addAction(about);
	init_deinit = new QAction("Connect", this);
	init_deinit->setCheckable(true);
	menu_device->addAction(init_deinit);
	menu_bar->addMenu(menu_file);
	menu_bar->addMenu(menu_device);
	menu_bar->addMenu(menu_help);
	Mwidget *m = new Mwidget(this);
	setCentralWidget(m);
	connect(m, SIGNAL(show_this(QString)), this, SLOT(show_message(QString)));
	connect(data_export, SIGNAL(triggered()), m, SLOT(save_data()));
	connect(about, SIGNAL(triggered()), this, SLOT(show_about()));
	connect(quit, SIGNAL(triggered()), qApp, SLOT(quit()));
	connect(init_deinit, SIGNAL(triggered(bool)), m, SLOT(init_deinit(bool)));
	QSettings settings("Surface Concept", "testApp");
	resize(settings.value("size/size", sizeHint()).toSize());
	move(settings.value("size/position", QPoint(200, 200)).toPoint());
}

void MainWindow::show_message(QString s)
{
	statusBar()->showMessage(s);
}

void MainWindow::closeEvent(QCloseEvent *ev)
{
	QSettings *s = new QSettings("Surface Concept", "testApp", this);
	s->setValue("size/size", size());
	s->setValue("size/position", pos());
	delete s;
	ev->accept();
}

void MainWindow::show_about()
{
	AboutDialog ad;
	ad.exec();
}
