#include "../../dld_dll_portable/src/scTDC.h"
#include "../../dld_dll_portable/src/scTDC_types.h"
#include <stdio.h>
#include <conio.h> // kbhit() getch()

int pipe_empty_returns = -1;

void eprint(int error)
{
	char msg[256];
	sc_get_err_msg(error, msg);
	printf("%s\n", msg);
}

void menu_print(void)
{
	printf("Options:\n");
	printf("\ts acq_time\tstart single measurement (acq_time - integer acquisition time [ms]. When 0 - infinite)\n");
	printf("\tr\t\trepeat single measurement (same acquisition time)\n");
	printf("\ta acq_time\tstart auto restart measurement (acq_time - integer acquisition time [ms])\n");
	printf("\tt\t\tstop auto restart measurement (running acquisition finishes normally)\n");
	printf("\ti\t\tinterrupt measurement (running acquisition is aborted)\n");
	printf("\tq\t\tquit\n");
	printf(">");
}

void turn_progress_bar(void)
{
	static unsigned char state = 0;
	state ++;
	state &= 0x3;
	switch(state)
	{
	case 0:
		printf("\b\b-");
		break;
	case 1:
		printf("\b\b\\");
		break;
	case 2:
		printf("\b\b|");
		break;
	case 3:
		printf("\b\b/");
		break;
	default: ;
	}
}

void callback(void *, int ret)
{
	eprint(ret);
	if(ret < 0) return;
	pipe_empty_returns = 0;
}

int main(int argc, char *argv[])
{
	bool output_to_file = false;
	bool output_to_console = false;
	FILE *output_file = NULL;
	char output_file_name[256];
	int buff_size = 0x400000;
	int ref_ch = -1;

// arguments parsing
	while(argc > 1)
	{
		if(argv[1][0] != '-')
		{
			printf("Wrong argument\n");
			return 0;
		}
		switch(argv[1][1])
		{
			case 'h':
				printf("Usage:\n");
				printf("\t-s buffer_size\t\tdefine buffer size\n");
				printf("\t-o output_file\t\tdefine output file name\n");
				printf("\t-c \t\t\tprint data to the console\n");
				printf("\t-r ref_channel\t\tdefine reference channel\n");
				return 0;
			case 'r': // ref channel
				if((argc < 3)||(argv[2][0] == '-')) // parameter is not given
				{
					printf("Parameter is missing\n");
					return 0;
				}
				if(sscanf(argv[2], "%d", &ref_ch) != 1) // cannot read parameter
				{
					printf("Wrong parameter\n");
					return 0;
				}
				printf("ref_channel = %d\n", ref_ch);
				++argv;
				--argc;
				break;
			case 'c':
				output_to_console = true;
				printf("Output to console is set\n");
				break;
			case 's': // buffer size
				if((argc < 3)||(argv[2][0] == '-')) // parameter is not given
				{
					printf("Parameter is missing\n");
					return 0;
				}
				if(sscanf(argv[2], "%d", &buff_size) != 1) // cannot read parameter
				{
					printf("Wrong parameter\n");
					return 0;
				}
				printf("buff_size = %d\n", buff_size);
				++argv;
				--argc;
				break;
			case 'o': // output file
				if((argc < 3)||(argv[2][0] == '-')) // parameter is not given
				{
					printf("Parameter is missing\n");
					return 0;
				}
				if(sscanf(argv[2], "%s", output_file_name) != 1) // cannot read parameter
				{
					printf("Wrong parameter\n");
					return 0;
				}
				printf("Output file: %s\n", output_file_name);
				output_file = fopen(output_file_name, "w");
				if(!output_file)
				{
					printf("Could not open file\n");
					return 0;
				}
				output_to_file = true;
				++argv;
				--argc;
				break;
			default:
				printf("Wrong parameter\n");
				return 0;
		}
		++argv;
		--argc;
	}
// end of parsing

	int ret;

// initialize
	ret = sc_tdc_init("./tdc_gpx3.ini");
	if(ret < 0)
	{
		printf("Initialize: ");
		eprint(ret);
		return ret;
	}

// get data format
	sc_tdc_format f;
	ret = sc_tdc_get_format(&f);
	if(ret < 0)
	{
		printf("Get format: ");
		eprint(ret);
		return ret;
	}
	printf("total bits len:%u\n", f.total_bits_length);
	printf("channel off:%u\n", f.channel_offset);
	printf("channel len:%u\n", f.channel_length);
	printf("time data off:%u\n", f.time_data_offset);
	printf("time data len:%u\n", f.time_data_length);
	printf("time tag off:%u\n", f.time_tag_offset);
	printf("time tag len:%u\n", f.time_tag_length);
	printf("start counter off:%u\n", f.start_counter_offset);
	printf("start counter len:%u\n", f.start_counter_length);
	printf("dif1 off:%u\n", f.dif1_offset);
	printf("dif1 len:%u\n", f.dif1_length);
	printf("dif2 off:%u\n", f.dif2_offset);
	printf("dif2 len:%u\n", f.dif2_length);
	printf("sum off:%u\n", f.sum_offset);
	printf("sum len:%u\n", f.sum_length);
	printf("flow_flags:%u\n", f.flow_control_flags);

// Set callback function
	ret = sc_tdc_set_complete_callback(0, callback);
	if(ret < 0)
	{
		printf("Set callback: ");
		eprint(ret);
		return ret;
	}

// Create pipe
	ret = sc_tdc_open_pipe(buff_size * (f.total_bits_length>>3));
	if(ret < 0)
	{
		printf("Open pipe: ");
		eprint(ret);
		return ret;
	}

// Get binsize
	double dData, bs;
	ret = sc_tdc_get_binsize(&bs);
	if(ret < 0)
	{
		printf("Get binsize: ");
		eprint(ret);
		return ret;
	}
	printf("binsize: %f ns\n", bs);

	unsigned int acq_time = 1;
	bool auto_restart = false;
	unsigned int acq_counter = 0;
	while(1)
	{
		if(auto_restart)
		{ // start new acquisition automatically
			pipe_empty_returns = -1;
			ret = sc_tdc_start_measure(acq_time);
			if(ret < 0)
			{
				printf("Start measurement:");
				eprint(ret);
				continue;
			}
		}
		else
		{ // ask user what to do
			char user_string[16];
			for(int i=0; i<sizeof(user_string); i++) user_string[i] = 0;
			char user_char;
			menu_print();
			gets(user_string);

			if(sscanf(user_string, "%c", &user_char) != 1)
			{
				printf("Unknown command: %s\n", user_string);
				continue;
			}

			if(user_char == 'q') break; // Exit

			auto_restart = false;
			switch(user_char)
			{
			case 'a': // auto restart mode
				auto_restart = true;
			case 's': // single acquisition mode
				if(sscanf(user_string, "%c %d", &user_char, &acq_time) != 2)
				{ // acquisition time is not given by the user
					auto_restart = false;
					printf("Missing parameter\n");
					continue;
				}
			case 'r': // single acquisition mode with the old acq_time
				pipe_empty_returns = -1;
				ret = sc_tdc_start_measure(acq_time);
				if(ret < 0)
				{
					auto_restart = false;
					printf("start measurement:");
					eprint(ret);
					continue;
				}
				break;
			default:
				{
					printf("Unknown command: %s\n", user_string);
					continue;
				}
			}
		}

		int buf_filling;
		unsigned int uiData[2];
		unsigned long long ullData;
		unsigned int uiTimeTag;
		unsigned int i = 0;
		unsigned int iter_counter = 0;
		unsigned int progress_bar = 0;
		bool print_data;
		unsigned int last_start_counter[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
		bool got_ref[16] = {false,false,false,false,false,false,false,false,false,false,false,false,false,false};
// print header
		if(f.dif1_length)
		{
			if(output_to_console)
			{
				printf("acquisition %d\n", acq_counter);
				printf("Iter\tCh\tStart\tData\tns\tDiff2\tDiff1\n");
			}
			if(output_to_file)
			{
				fprintf(output_file, "acquisition %d\n", acq_counter);
				fprintf(output_file, "Iter\tCh\tStart\tData\tns\tDiff2\tDiff1\n");
			}
		}
		else
		{
			if(f.total_bits_length == 64)
			{
				if(output_to_console)
				{
					printf("acquisition %d\n", acq_counter);
					printf("Iter\tCh\tStart\tData\tns\tTag\n");
				}
				if(output_to_file)
				{
					fprintf(output_file, "acquisition %d\n", acq_counter);
					fprintf(output_file, "Iter\tCh\tStart\tData\tns\tTag\n");
				}
			}
			else
			{
				if(output_to_console)
				{
					printf("acquisition %d\n", acq_counter);
					printf("Iter\tCh\tStart\tData\tns\n");
				}
				if(output_to_file)
				{
					fprintf(output_file, "acquisition %d\n", acq_counter);
					fprintf(output_file, "Iter\tCh\tStart\tData\tns\n");
				}
			}
		}
// end of header
		do
		{ // extract and print data
			if(f.total_bits_length == 64)
			{ // 64 bits mode
				buf_filling = sc_tdc_get_next_data((void *)uiData);
				if(buf_filling >= 0)
				{
					ullData = *((unsigned long long *)uiData);
					if(ullData == 0xFFFFFFFFFFFFFFFB) // extract statistics from the data flow
					{
						for(i=0; i<(sizeof(statistics_t)>>3); i++)
						{
							buf_filling = sc_tdc_get_next_data((void *)uiData);
							if(buf_filling < 0) i--;
							else if(output_to_console)
							{
							
								switch(i)
								{
									case 0:
									case 2:
									case 4:
									case 8:
									case 10:
									case 12:
										printf("%d\t", uiData[0]);
										break;
									case 6:
									case 14:
										printf("%d\n", uiData[0]);
										break;
									case 64:
										printf("%d\t%d\t", uiData[0], uiData[1]);
										break;
									case 66:
										printf("%d\t%d\n", uiData[0], uiData[1]);
										break;
									default:
										break;
								}
							}
						}
					}
					else if(ullData == 0xFFFFFFFFFFFFFFFA) // ms tick
					{
						progress_bar ++;
					}
					else
					{ // print data
						unsigned int channel, start_counter, diff1, diff2;
						unsigned int time_data[2];
						unsigned long long *ullp_time_data = (unsigned long long *)(&time_data);
						
						if(f.dif1_length)
						{ // dld mode
							channel = (ullData >> f.channel_offset) & sc_mask64[f.channel_length];
							start_counter = (ullData >> f.start_counter_offset) & sc_mask64[f.start_counter_length];
							*ullp_time_data = (ullData >> f.sum_offset) & sc_mask64[f.sum_length];
							diff2 = (ullData >> f.dif2_offset) & sc_mask64[f.dif2_length];
							diff1 = (ullData >> f.dif1_offset) & sc_mask64[f.dif1_length];
							dData = (*ullp_time_data) * bs;
							if(output_to_console)
								printf("%d\t%d\t%d\t%08x%08x\t%f\t%d\t%d\n", iter_counter, channel, start_counter, time_data[1], time_data[0], dData,  diff2, diff1);
							if(output_to_file)
								fprintf(output_file, "%d\t%d\t%d\t%08x%08x\t%f\t%d\t%d\n", iter_counter, channel, start_counter, time_data[1], time_data[0], dData,  diff2, diff1);
						}
						else
						{ // tdc mode
							channel = (ullData >> f.channel_offset) & sc_mask64[f.channel_length];
							start_counter = (ullData >> f.start_counter_offset) & sc_mask64[f.start_counter_length];
							uiTimeTag = (ullData >> f.time_tag_offset) & sc_mask64[f.time_tag_length];
							*ullp_time_data = (ullData >> f.time_data_offset) & sc_mask64[f.time_data_length];
							dData = (*ullp_time_data) * bs;
							print_data = true;
							if(ref_ch >= 0)
							{
								if(channel == ref_ch)
								{
									print_data = false;
									for(unsigned int ch=0; ch<16; ch++) got_ref[ch] = true;
								}
								else if(got_ref[channel]) got_ref[channel] = false;
								else print_data = false;
							}
							if(print_data)
							{
								if(output_to_console)
									printf("%d\t%d\t%d\t%08x%08x\t%f\t%u\n", iter_counter, channel, start_counter, time_data[1], time_data[0], dData, uiTimeTag);
								if(output_to_file)
									fprintf(output_file, "%d\t%d\t%d\t%08x%08x\t%f\t%u\n", iter_counter, channel, start_counter, time_data[1], time_data[0], dData, uiTimeTag);
							}
						}
						if(((last_start_counter[channel] + 1) & sc_mask32[f.start_counter_length]) != start_counter)
						{
//							if(output_to_console)
								printf("START COUNTER ERROR!!!\n");
							if(output_to_file)
								fprintf(output_file, "START COUNTER ERROR!!!\n");
						}
						last_start_counter[channel] = start_counter;
					}
					if(pipe_empty_returns > 0) pipe_empty_returns = 0;
				}
				else
				{
					progress_bar ++;
					if(pipe_empty_returns >= 0) pipe_empty_returns ++;
				}
			}
			else
			{ // 32 bits mode
				buf_filling = sc_tdc_get_next_data((void *)uiData);
				if(buf_filling >= 0)
				{
					if(uiData[0] == 0xFFFFFFFB) // extract statistics from the data flow
					{
						for(i=0; i<(sizeof(statistics_t)>>2); i++)
						{
							buf_filling = sc_tdc_get_next_data((void *)uiData);
							if(buf_filling < 0) i--;
						}
					}
					if(uiData[0] == 0xFFFFFFFA) // ms tick
					{
						progress_bar ++;
					}
					else
					{ // print data
						unsigned int channel, start_counter, time_data, diff1, diff2;
						if(f.dif1_length)
						{ // dld mode
							channel = (uiData[0] >> f.channel_offset) & sc_mask32[f.channel_length];
							start_counter = (uiData[0] >> f.start_counter_offset) & sc_mask32[f.start_counter_length];
							time_data = (uiData[0] >> f.sum_offset) & sc_mask32[f.sum_length];
							diff2 = (uiData[0] >> f.dif2_offset) & sc_mask32[f.dif2_length];
							diff1 = (uiData[0] >> f.dif1_offset) & sc_mask32[f.dif1_length];
							dData = time_data * bs;
							if(output_to_console)
								printf("%d\t%d\t%d\t%d\t%f\t%d\t%d\n", iter_counter, channel, start_counter, time_data, dData, diff2, diff1);
							if(output_to_file)
								fprintf(output_file, "%d\t%d\t%d\t%d\t%f\t%d\t%d\n", iter_counter, channel, start_counter, time_data, dData, diff2, diff1);
						}
						else
						{ // tdc mode
							channel = (uiData[0] >> f.channel_offset) & sc_mask32[f.channel_length];
							start_counter = (uiData[0] >> f.start_counter_offset) & sc_mask32[f.start_counter_length];
							time_data = (uiData[0] >> f.time_data_offset) & sc_mask32[f.time_data_length];
							dData = time_data * bs;
							print_data = true;
							if(ref_ch >= 0)
							{
								if(channel == ref_ch)
								{
									print_data = false;
									for(unsigned int ch=0; ch<16; ch++) got_ref[ch] = true;
								}
								else if(got_ref[channel]) got_ref[channel] = false;
								else print_data = false;
							}
							if(print_data)
							{
								if(output_to_console)
									printf("%d\t%d\t%d\t%08x\t%f\n", iter_counter, channel, start_counter, time_data, dData);
								if(output_to_file)
									fprintf(output_file, "%d\t%d\t%d\t%08x\t%f\n", iter_counter, channel, start_counter, time_data, dData);
							}
						}
						if(((last_start_counter[channel] + 1) & sc_mask32[f.start_counter_length]) != start_counter)
						{
//							if(output_to_console)
								printf("START COUNTER ERROR!!!\n");
							if(output_to_file)
								fprintf(output_file, "START COUNTER ERROR!!!\n");
						}
						last_start_counter[channel] = start_counter;
					}
					if(pipe_empty_returns > 0) pipe_empty_returns = 0;
				}
				else
				{
					progress_bar ++;
					if(pipe_empty_returns >= 0) pipe_empty_returns ++;
				}
			}
			if(progress_bar > 100)
			{
				if(!output_to_console)
				{ // progress stick turning
					turn_progress_bar();
				}
				if(kbhit())
				{ // user has pressed a key
					switch(getch())
					{
					case 'i': // interrupt
						ret = sc_tdc_interrupt();
						if(ret < 0)
						{
							printf("interrupt: ");
							eprint(ret);
							return ret;
						}
					case 't': // autorestart off
						auto_restart = false;
						break;
					default : ;
					}
				}
				progress_bar = 0;
			}
		} while((pipe_empty_returns < 0) || (buf_filling >= 0));
		acq_counter ++;
	}

	if(output_file) fclose(output_file);

	ret = sc_tdc_close_pipe();
	if(ret < 0)
	{
		printf("close pipe: ");
		eprint(ret);
		return ret;
	}
	
	ret = sc_tdc_deinit();
	if(ret < 0)
	{
		printf("deinit: ");
		eprint(ret);
		return ret;
	}
	return 0;
}

