#include "./scTDC.h"
#include "./scTDC_types.h"
#include "./scTDC_error_codes.h"
#include <windows.h>
#include <stdio.h>
#include <conio.h> // kbhit() getch()
#include "iniparser.h"
#include <string>
#include <sstream>
#include <iomanip>
#include <iostream>
using std::string;
using std::stringstream;


#define STREAM_PIPE 0x02
#define BUFFER_PIPE 0x04
#define INI_FILE_NAME "pipelist.ini"

unsigned int pipe_count;
sc_pipe_type_t pipe_types[16];
void *pipe_params[16];
int pipe_ids[16];
int device_id;

int pipe_empty_returns = -1;
roi_t roi;
unsigned long long histo_size;
unsigned int stream_pipe_size;
std::string output_file_name;

sc3du_t hw_bin;

//===================================================================

void eprint(int error)
{
	char msg[256];
	sc_get_err_msg(error, msg);
	printf("%s\n", msg);
}

//===================================================================

void menu_print(void)
{
	printf("Options:\n");
	printf("\ts acq_time\tstart single acquisition (acq_time - integer acquisition time [ms]. When 0 - infinite)\n");
	printf("\tr\t\trepeat single acquisition (same acquisition time)\n");
	printf("\ta acq_time\tstart auto restart acquisition (acq_time - integer acquisition time [ms])\n");
	printf("\tt\t\tstop auto restart acquisition (running acquisition finishes normally)\n");
	printf("\ti\t\tinterrupt acquisition (running acquisition is aborted)\n");
	printf("\tq\t\tquit\n");
	printf(">");
}

//===================================================================

int read_ini_file(const char *filename)
{
	string inistr, section, section_base, parameter;
	dictionary* dict = NULL;
	dict = iniparser_load(filename);
	if(!dict)
		return -1;

	pipe_count = (unsigned int)iniparser_getint(dict, "COMMON:pipe_count", 0);
	if(pipe_count > 16) return -1;

	output_file_name = iniparser_getstring(dict, "COMMON:output_file_name_base", "output");

	stream_pipe_size = iniparser_getint(dict, "STREAM_PIPE:pipe_size", 0x100000);
	
	section_base = "PIPE";

	for(unsigned int i=0; i<pipe_count; i++)
	{
		std::ostringstream convert;
		convert << std::setw(2) << std::setfill('0') << i;
		section = section_base + convert.str();
		inistr = iniparser_getstring(dict, (section + ":type").c_str(), "no_type");
		if(!inistr.compare("image_xy"))
		{
			pipe_types[i] = DLD_IMAGE_XY;
			pipe_params[i] = (void *) new sc_pipe_dld_image_xy_params_t;
			switch(iniparser_getint(dict, (section + ":depth").c_str(), 32))
			{
			case 8:
				((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->depth = BS8;
				break;
			case 16:
				((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->depth = BS16;
				break;
			case 32:
			default:
				((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->depth = BS32;
				break;
			case 64:
				((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->depth = BS64;
				break;
			}
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				modulo = iniparser_getint(dict, (section + ":modulo").c_str(), 0);
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				binning.x = iniparser_getint(dict, (section + ":binning_x").c_str(), 1);
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				binning.y = iniparser_getint(dict, (section + ":binning_y").c_str(), 1);
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				binning.time = iniparser_getint(dict, (section + ":binning_time").c_str(), 1);
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				roi.offset.x = iniparser_getint(dict, (section + ":roi_offset_x").c_str(), 0);
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				roi.offset.y = iniparser_getint(dict, (section + ":roi_offset_y").c_str(), 0);
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				roi.offset.time = iniparser_getint(dict, (section + ":roi_offset_time").c_str(), 0);
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				roi.size.x = iniparser_getint(dict, (section + ":roi_size_x").c_str(), 512);
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				roi.size.y = iniparser_getint(dict, (section + ":roi_size_y").c_str(), 512);
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				roi.size.time = iniparser_getint(dict, (section + ":roi_size_time").c_str(), 65536);
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				accumulation_ms = iniparser_getint(dict, (section + ":accu_time_ms").c_str(), 0);
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				channel = iniparser_getint(dict, (section + ":channel").c_str(), -1);
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				allocator_owner = NULL;
			((sc_pipe_dld_image_xy_params_t *)pipe_params[i])->
				allocator_cb = NULL;
		}
		else if(!inistr.compare("image_xt"))
		{
			pipe_types[i] = DLD_IMAGE_XT;
			pipe_params[i] = (void *) new sc_pipe_dld_image_xt_params_t;
			switch(iniparser_getint(dict, (section + ":depth").c_str(), 32))
			{
			case 8:
				((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->depth = BS8;
				break;
			case 16:
				((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->depth = BS16;
				break;
			case 32:
			default:
				((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->depth = BS32;
				break;
			case 64:
				((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->depth = BS64;
				break;
			}
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				modulo = iniparser_getint(dict, (section + ":modulo").c_str(), 0);
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				binning.x = iniparser_getint(dict, (section + ":binning_x").c_str(), 1);
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				binning.y = iniparser_getint(dict, (section + ":binning_y").c_str(), 1);
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				binning.time = iniparser_getint(dict, (section + ":binning_time").c_str(), 1);
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				roi.offset.x = iniparser_getint(dict, (section + ":roi_offset_x").c_str(), 0);
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				roi.offset.y = iniparser_getint(dict, (section + ":roi_offset_y").c_str(), 0);
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				roi.offset.time = iniparser_getint(dict, (section + ":roi_offset_time").c_str(), 0);
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				roi.size.x = iniparser_getint(dict, (section + ":roi_size_x").c_str(), 512);
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				roi.size.y = iniparser_getint(dict, (section + ":roi_size_y").c_str(), 512);
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				roi.size.time = iniparser_getint(dict, (section + ":roi_size_time").c_str(), 65536);
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				accumulation_ms = iniparser_getint(dict, (section + ":accu_time_ms").c_str(), 0);
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				channel = iniparser_getint(dict, (section + ":channel").c_str(), -1);
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				allocator_owner = NULL;
			((sc_pipe_dld_image_xt_params_t *)pipe_params[i])->
				allocator_cb = NULL;
		}
		else if(!inistr.compare("image_yt"))
		{
			pipe_types[i] = DLD_IMAGE_YT;
			pipe_params[i] = (void *) new sc_pipe_dld_image_yt_params_t;
			switch(iniparser_getint(dict, (section + ":depth").c_str(), 32))
			{
			case 8:
				((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->depth = BS8;
				break;
			case 16:
				((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->depth = BS16;
				break;
			case 32:
			default:
				((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->depth = BS32;
				break;
			case 64:
				((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->depth = BS64;
				break;
			}
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				modulo = iniparser_getint(dict, (section + ":modulo").c_str(), 0);
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				binning.x = iniparser_getint(dict, (section + ":binning_x").c_str(), 1);
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				binning.y = iniparser_getint(dict, (section + ":binning_y").c_str(), 1);
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				binning.time = iniparser_getint(dict, (section + ":binning_time").c_str(), 1);
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				roi.offset.x = iniparser_getint(dict, (section + ":roi_offset_x").c_str(), 0);
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				roi.offset.y = iniparser_getint(dict, (section + ":roi_offset_y").c_str(), 0);
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				roi.offset.time = iniparser_getint(dict, (section + ":roi_offset_time").c_str(), 0);
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				roi.size.x = iniparser_getint(dict, (section + ":roi_size_x").c_str(), 512);
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				roi.size.y = iniparser_getint(dict, (section + ":roi_size_y").c_str(), 512);
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				roi.size.time = iniparser_getint(dict, (section + ":roi_size_time").c_str(), 65536);
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				accumulation_ms = iniparser_getint(dict, (section + ":accu_time_ms").c_str(), 0);
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				channel = iniparser_getint(dict, (section + ":channel").c_str(), -1);
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				allocator_owner = NULL;
			((sc_pipe_dld_image_yt_params_t *)pipe_params[i])->
				allocator_cb = NULL;
		}
		else if(!inistr.compare("image_3d"))
		{
			pipe_types[i] = DLD_IMAGE_3D;
			pipe_params[i] = (void *) new sc_pipe_dld_image_3d_params_t;
			switch(iniparser_getint(dict, (section + ":depth").c_str(), 32))
			{
			case 8:
				((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->depth = BS8;
				break;
			case 16:
				((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->depth = BS16;
				break;
			case 32:
			default:
				((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->depth = BS32;
				break;
			case 64:
				((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->depth = BS64;
				break;
			}
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				modulo = iniparser_getint(dict, (section + ":modulo").c_str(), 0);
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				binning.x = iniparser_getint(dict, (section + ":binning_x").c_str(), 1);
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				binning.y = iniparser_getint(dict, (section + ":binning_y").c_str(), 1);
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				binning.time = iniparser_getint(dict, (section + ":binning_time").c_str(), 1);
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				roi.offset.x = iniparser_getint(dict, (section + ":roi_offset_x").c_str(), 0);
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				roi.offset.y = iniparser_getint(dict, (section + ":roi_offset_y").c_str(), 0);
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				roi.offset.time = iniparser_getint(dict, (section + ":roi_offset_time").c_str(), 0);
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				roi.size.x = iniparser_getint(dict, (section + ":roi_size_x").c_str(), 512);
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				roi.size.y = iniparser_getint(dict, (section + ":roi_size_y").c_str(), 512);
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				roi.size.time = iniparser_getint(dict, (section + ":roi_size_time").c_str(), 65536);
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				accumulation_ms = iniparser_getint(dict, (section + ":accu_time_ms").c_str(), 0);
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				channel = iniparser_getint(dict, (section + ":channel").c_str(), -1);
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				allocator_owner = NULL;
			((sc_pipe_dld_image_3d_params_t *)pipe_params[i])->
				allocator_cb = NULL;
		}
		else if(!inistr.compare("sum_histo"))
		{
			pipe_types[i] = DLD_SUM_HISTO;
			pipe_params[i] = (void *) new sc_pipe_dld_sum_histo_params_t;
			switch(iniparser_getint(dict, (section + ":depth").c_str(), 32))
			{
			case 8:
				((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->depth = BS8;
				break;
			case 16:
				((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->depth = BS16;
				break;
			case 32:
			default:
				((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->depth = BS32;
				break;
			case 64:
				((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->depth = BS64;
				break;
			}
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				modulo = iniparser_getint(dict, (section + ":modulo").c_str(), 0);
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				binning.x = iniparser_getint(dict, (section + ":binning_x").c_str(), 1);
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				binning.y = iniparser_getint(dict, (section + ":binning_y").c_str(), 1);
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				binning.time = iniparser_getint(dict, (section + ":binning_time").c_str(), 1);
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				roi.offset.x = iniparser_getint(dict, (section + ":roi_offset_x").c_str(), 0);
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				roi.offset.y = iniparser_getint(dict, (section + ":roi_offset_y").c_str(), 0);
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				roi.offset.time = iniparser_getint(dict, (section + ":roi_offset_time").c_str(), 0);
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				roi.size.x = iniparser_getint(dict, (section + ":roi_size_x").c_str(), 512);
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				roi.size.y = iniparser_getint(dict, (section + ":roi_size_y").c_str(), 512);
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				roi.size.time = iniparser_getint(dict, (section + ":roi_size_time").c_str(), 65536);
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				accumulation_ms = iniparser_getint(dict, (section + ":accu_time_ms").c_str(), 0);
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				channel = iniparser_getint(dict, (section + ":channel").c_str(), -1);
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				allocator_owner = NULL;
			((sc_pipe_dld_sum_histo_params_t *)pipe_params[i])->
				allocator_cb = NULL;
		}
		else if(!inistr.compare("tdc_histo"))
		{
			pipe_types[i] = TDC_HISTO;
			pipe_params[i] = (void *) new sc_pipe_tdc_histo_params_t;
			switch(iniparser_getint(dict, (section + ":depth").c_str(), 32))
			{
			case 8:
				((sc_pipe_tdc_histo_params_t *)pipe_params[i])->depth = BS8;
				break;
			case 16:
				((sc_pipe_tdc_histo_params_t *)pipe_params[i])->depth = BS16;
				break;
			case 32:
			default:
				((sc_pipe_tdc_histo_params_t *)pipe_params[i])->depth = BS32;
				break;
			case 64:
				((sc_pipe_tdc_histo_params_t *)pipe_params[i])->depth = BS64;
				break;
			}
			((sc_pipe_tdc_histo_params_t *)pipe_params[i])->
				modulo = iniparser_getint(dict, (section + ":modulo").c_str(), 0);
			((sc_pipe_tdc_histo_params_t *)pipe_params[i])->
				channel = iniparser_getint(dict, (section + ":channel").c_str(), -1);
			((sc_pipe_tdc_histo_params_t *)pipe_params[i])->
				binning = iniparser_getint(dict, (section + ":binning").c_str(), 1);
			((sc_pipe_tdc_histo_params_t *)pipe_params[i])->
				offset = iniparser_getint(dict, (section + ":offset").c_str(), 0);
			((sc_pipe_tdc_histo_params_t *)pipe_params[i])->
				size = iniparser_getint(dict, (section + ":size").c_str(), 0x10000);
			((sc_pipe_tdc_histo_params_t *)pipe_params[i])->
				accumulation_ms = iniparser_getint(dict, (section + ":accu_time_ms").c_str(), 0);
			((sc_pipe_tdc_histo_params_t *)pipe_params[i])->
				allocator_owner = NULL;
			((sc_pipe_tdc_histo_params_t *)pipe_params[i])->
				allocator_cb = NULL;
		}
		else if(!inistr.compare("statistics"))
		{
			pipe_types[i] = STATISTICS;
			pipe_params[i] = (void *) new sc_pipe_statistics_params_t;
			((sc_pipe_statistics_params_t *)pipe_params[i])->
				allocator_owner = NULL;
			((sc_pipe_statistics_params_t *)pipe_params[i])->
				allocator_cb = NULL;
		}
		else
			return -1;
	}


	hw_bin.x = iniparser_getint(dict, "DLD:hw_binning_x", 1);
	hw_bin.y = iniparser_getint(dict, "DLD:hw_binning_y", 1);
	hw_bin.time = iniparser_getint(dict, "DLD:hw_binning_time", 1);

	iniparser_freedict(dict);
	return 0;
}

//===================================================================

int subscribe(unsigned int subscribe_mask)
{
	int ret = 0;

// Subscribe to STREAM_PIPE
	if(subscribe_mask & STREAM_PIPE)
	{
		ret = sc_tdc_pipe_open2(device_id, stream_pipe_size, NULL, NULL);
		if(ret < 0)
		{
			printf("Open stream pipe: ");
			eprint(ret);
			return ret;
		}
	}
// subscribe to BUFFER_PIPE
	if(subscribe_mask & BUFFER_PIPE)
	{
		for(int i=0; i<pipe_count; i++)
		{
			pipe_ids[i] = sc_pipe_open2(device_id, pipe_types[i], pipe_params[i]);
			if(pipe_ids[i] < 0)
			{
				printf("Open pipe: ");
				eprint(pipe_ids[i]);
				return pipe_ids[i];
			}
		}
	}
	return 0;
}

//===================================================================

int unsubscribe(unsigned int subscribe_mask)
{
	int ret = 0;

// unsubscribe from STREAM_PIPE
	if(subscribe_mask & STREAM_PIPE)
	{
		ret = sc_tdc_pipe_close2(device_id);
		if(ret < 0)
		{
			printf("Close stream pipe: ");
			eprint(ret);
			return ret;
		}
	}

// unsubscribe from BUFFER_PIPE
	if(subscribe_mask & BUFFER_PIPE)
	{
		for(int i=0; i<pipe_count; i++)
		{
			ret = sc_pipe_close2(device_id, pipe_ids[i]);
			if(ret < 0)
			{
				printf("Close tdc histo pipe: ");
				eprint(ret);
				return ret;
			}
		}
	}
	return 0;
}

//===================================================================

void turn_progress_bar(void)
{
	static unsigned char state = 0;
	state ++;
	state &= 0x3;
	switch(state)
	{
	case 0:
		printf("\b\b-");
		break;
	case 1:
		printf("\b\b\\");
		break;
	case 2:
		printf("\b\b|");
		break;
	case 3:
		printf("\b\b/");
		break;
	default: ;
	}
}

//===================================================================

void callback(void *, int ret)
{
	eprint(ret);
	if(ret < 0) return;
	pipe_empty_returns = 0;
}

//===================================================================

int main(int argc, char *argv[])
{
	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);//	bool output_to_file = false;
	bool output_to_console = false;
	FILE *logfile = fopen("pipe_read.log", "w");
	FILE *output_file = NULL;
	FILE *pipe_buf_output_file[16];
	char subscribe_list[32];
	unsigned int subscribe_mask = 0;

// arguments parsing
	while(argc > 1)
	{
		if(argv[1][0] != '-')
		{
			printf("Wrong argument\n");
			return 0;
		}
		unsigned int ch = 0;
		switch(argv[1][1])
		{
		case 'h':
			printf("Usage:\n");
			printf("\t-s subscribe list\t\t[P|I]\n");
			printf("\t-c \t\t\tprint data to the console\n");
			return 0;
		case 'c':
			output_to_console = true;
			printf("Output to console is set\n");
			break;
		case 's':
			if((argc < 3)||(argv[2][0] == '-')) // parameter is not given
			{
				printf("Parameter is missing\n");
				return 0;
			}
			if(sscanf(argv[2], "%s", subscribe_list) != 1) // cannot read parameter
			{
				printf("Wrong parameter\n");
				return 0;
			}
			subscribe_mask = 0;
			while(subscribe_list[ch] != 0)
			{
				switch(subscribe_list[ch])
				{
				case 'P':
					subscribe_mask |= STREAM_PIPE;
					break;
				case 'I':
					subscribe_mask |= BUFFER_PIPE;
					break;
				default:
					printf("Wrong parameter\n");
					return 0;
				}
				ch++;
			}
			++argv;
			--argc;
			break;
		default:
			printf("Wrong parameter\n");
			return 0;
		}
		++argv;
		--argc;
	}
// end of parsing

	int ret;

// read ini file
	ret = read_ini_file(INI_FILE_NAME);
	if(ret < 0)
	{
		printf("Cannot read ini file\n");
		return ret;
	}

// initialize tdc
	device_id = sc_tdc_init_inifile("./tdc_gpx3.ini");
//	device_id = sc_tdc_init("./tdc_gpx3.ini");
	if(device_id < 0)
	{
		printf("Initialize: ");
		eprint(device_id);
		return device_id;
	}


// subscribe
	ret = subscribe(subscribe_mask);
	if(ret < 0)
	{
		printf("Subscribing has failed\n");
		return ret;
	}

// get data format
	sc_tdc_format f;
	ret = sc_tdc_get_format2(device_id, &f);
	if(ret < 0)
	{
		printf("Get format: ");
		eprint(ret);
		return ret;
	}
	printf("total bits len:%u\n", f.total_bits_length);
	printf("channel off:%u\n", f.channel_offset);
	printf("channel len:%u\n", f.channel_length);
	printf("time data off:%u\n", f.time_data_offset);
	printf("time data len:%u\n", f.time_data_length);
	printf("time tag off:%u\n", f.time_tag_offset);
	printf("time tag len:%u\n", f.time_tag_length);
	printf("start counter off:%u\n", f.start_counter_offset);
	printf("start counter len:%u\n", f.start_counter_length);
	printf("sign counter off:%u\n", f.sign_counter_offset);
	printf("sign counter len:%u\n", f.sign_counter_length);
	printf("dif1 off:%u\n", f.dif1_offset);
	printf("dif1 len:%u\n", f.dif1_length);
	printf("dif2 off:%u\n", f.dif2_offset);
	printf("dif2 len:%u\n", f.dif2_length);
	printf("sum off:%u\n", f.sum_offset);
	printf("sum len:%u\n", f.sum_length);
	printf("flow_flags:%u\n", f.flow_control_flags);

// Set callback function
	ret = sc_tdc_set_complete_callback2(device_id, 0, callback);
	if(ret < 0)
	{
		printf("Set callback: ");
		eprint(ret);
		return ret;
	}

// Get binsize
	double dData, bs;
	ret = sc_tdc_get_binsize2(device_id, &bs);
	if(ret < 0)
	{
		printf("Get binsize: ");
		eprint(ret);
		return ret;
	}
	printf("binsize: %f ns\n", bs);

	bool histo_changed = false;

	unsigned int acq_time = 1;
	bool auto_restart = false;
	unsigned int acq_counter = 0;
	unsigned int accu_counter[16];

	while(1)
	{
		if(auto_restart)
		{ // start new acquisition automatically
			pipe_empty_returns = -1;
			ret = sc_tdc_start_measure2(device_id, acq_time);
			if(ret < 0)
			{
				printf("Start acquisition:");
				eprint(ret);
				continue;
			}
		}
		else
		{ // ask user what to do
			char user_string[16];
			for(int i=0; i<sizeof(user_string); i++) user_string[i] = 0;
			char user_char;
			menu_print();
			gets(user_string);

			if(sscanf(user_string, "%c", &user_char) != 1)
			{
				printf("Unknown command: %s\n", user_string);
				continue;
			}

			if(user_char == 'q') break; // Exit

			auto_restart = false;
			switch(user_char)
			{
			case 'c': // reconfigure
			// unsubscribe
				ret = unsubscribe(subscribe_mask);
				if(ret < 0)
				{
					printf("Unubscribing has failed\n");
					return ret;
				}
			// read ini file
				ret = read_ini_file(INI_FILE_NAME);
				if(ret < 0)
				{
					printf("Cannot read ini file\n");
					return ret;
				}
				
			// subscribe
				ret = subscribe(subscribe_mask);
				if(ret < 0)
				{
					printf("Subscribing has failed\n");
					return ret;
				}
				continue;
			case 'a': // auto restart mode
				auto_restart = true;
			case 's': // single acquisition mode
				if(sscanf(user_string, "%c %d", &user_char, &acq_time) != 2)
				{ // acquisition time is not given by the user
					auto_restart = false;
					printf("Missing parameter\n");
					continue;
				}
			case 'r': // single acquisition mode with the old acq_time
				pipe_empty_returns = -1;
				ret = sc_dld_set_hardware_binning(device_id, &hw_bin);
				if(ret < 0)
				{
					printf("set hw binning:");
					eprint(ret);
					continue;
				}
				ret = sc_tdc_start_measure2(device_id, acq_time);
				if(ret < 0)
				{
					auto_restart = false;
					printf("start acquisition:");
					eprint(ret);
					continue;
				}
				break;
			default:
				{
					printf("Unknown command: %s\n", user_string);
					continue;
				}
			}
		}
		unsigned int iter_counter = 0;

		if(subscribe_mask & STREAM_PIPE)
		{
			FILE *output_file = fopen((output_file_name + "_stream_pipe.dat").c_str(), "w");

			int buf_filling;
			unsigned int uiData[2];
			unsigned long long ullData;
			unsigned int uiTimeTag;
			unsigned int i = 0;
			unsigned int progress_bar = 0;
	// print header
			if(f.dif1_length)
			{
				if(output_to_console)
				{
					printf("acquisition %d\n", acq_counter);
					printf("Iter\tCh\tStart\tData\tns\tDiff2\tDiff1\n");
				}
			//	if(output_to_file)
				{
					fprintf(output_file, "acquisition %d\n", acq_counter);
					fprintf(output_file, "Iter\tCh\tStart\tData\tns\tDiff2\tDiff1\n");
				}
			}
			else
			{
				if(f.total_bits_length == 64)
				{
					if(output_to_console)
					{
						printf("acquisition %d\n", acq_counter);
						printf("Iter\tCh\tSign\tStart\tData\tns\tTag\n");
					}
				//	if(output_to_file)
					{
						fprintf(output_file, "acquisition %d\n", acq_counter);
						fprintf(output_file, "Iter\tCh\tSign\tStart\tData\tns\tTag\n");
					}
				}
				else
				{
					if(output_to_console)
					{
						printf("acquisition %d\n", acq_counter);
						printf("Iter\tCh\tSign\tStart\tData\tns\n");
					}
				//	if(output_to_file)
					{
						fprintf(output_file, "acquisition %d\n", acq_counter);
						fprintf(output_file, "Iter\tCh\tSign\tStart\tData\tns\n");
					}
				}
			}
	// end of header
			do
			{ // extract and print data
				if(f.total_bits_length == 64)
				{ // 64 bits mode
					buf_filling = sc_tdc_pipe_read2(device_id, (void *)uiData, 8, 10);
					if(buf_filling > 0)
					{
						ullData = *((unsigned long long *)uiData);
						if(ullData == 0xFFFFFFFFFFFFFFFB) // extract statistics from the data flow
						{
							for(i=0; i<(sizeof(statistics_t)>>3); i++)
							{
								buf_filling = sc_tdc_pipe_read2(device_id, (void *)uiData, 8, -1);
								if(buf_filling == 0) i--;
							}
						}
						else if(ullData == 0xFFFFFFFFFFFFFFFA) // ms tick
						{
							progress_bar ++;
						}
						else
						{ // print data
							unsigned int channel, start_counter, sign_counter, diff1, diff2;
							unsigned int time_data[2];
							unsigned long long *ullp_time_data = (unsigned long long *)(&time_data);
							
							if(f.dif1_length)
							{ // dld mode
								channel = (ullData >> f.channel_offset) & sc_mask64[f.channel_length];
								start_counter = (ullData >> f.start_counter_offset) & sc_mask64[f.start_counter_length];
								*ullp_time_data = (ullData >> f.sum_offset) & sc_mask64[f.sum_length];
								diff2 = (ullData >> f.dif2_offset) & sc_mask64[f.dif2_length];
								diff1 = (ullData >> f.dif1_offset) & sc_mask64[f.dif1_length];
								dData = (*ullp_time_data) * bs;
								if(output_to_console)
									printf("%d\t%d\t%d\t%08x%08x\t%f\t%d\t%d\n", iter_counter, channel, start_counter, time_data[1], time_data[0], dData,  diff2, diff1);
							//	if(output_to_file)
									fprintf(output_file, "%d\t%d\t%d\t%08x%08x\t%f\t%d\t%d\n", iter_counter, channel, start_counter, time_data[1], time_data[0], dData,  diff2, diff1);
							}
							else
							{ // tdc mode
								channel = (ullData >> f.channel_offset) & sc_mask64[f.channel_length];
								start_counter = (ullData >> f.start_counter_offset) & sc_mask64[f.start_counter_length];
								sign_counter = (ullData >> f.sign_counter_offset) & sc_mask64[f.sign_counter_length];
								uiTimeTag = (ullData >> f.time_tag_offset) & sc_mask64[f.time_tag_length];
								*ullp_time_data = (ullData >> f.time_data_offset) & sc_mask64[f.time_data_length];
								dData = (*ullp_time_data) * bs;
								if(output_to_console)
									printf("%d\t%d\t%d\t%d\t%08x%08x\t%f\t%u\n", iter_counter, channel, sign_counter, start_counter, time_data[1], time_data[0], dData, uiTimeTag);
							//	if(output_to_file)
									fprintf(output_file, "%d\t%d\t%d\t%d\t%08x%08x\t%f\t%u\n", iter_counter, channel, sign_counter, start_counter, time_data[1], time_data[0], dData, uiTimeTag);
							}
						}
						if(pipe_empty_returns > 0) pipe_empty_returns = 0;
					}
					else
					{
						progress_bar ++;
						if(pipe_empty_returns >= 0) pipe_empty_returns ++;
					}
				}
				else
				{ // 32 bits mode
					buf_filling = sc_tdc_pipe_read2(device_id, (void *)uiData, 4, 10);
					if(buf_filling > 0)
					{
						if(uiData[0] == 0xFFFFFFFB) // extract statistics from the data flow
						{
							for(i=0; i<(sizeof(statistics_t)>>2); i++)
							{
								buf_filling = sc_tdc_pipe_read2(device_id, (void *)uiData, 4, -1);
								if(buf_filling == 0) i--;
							}
						}
						if(uiData[0] == 0xFFFFFFFA) // ms tick
						{
							progress_bar ++;
						}
						else
						{ // print data
							unsigned int channel, start_counter, sign_counter, time_data, diff1, diff2;
							if(f.dif1_length)
							{ // dld mode
								channel = (uiData[0] >> f.channel_offset) & sc_mask32[f.channel_length];
								start_counter = (uiData[0] >> f.start_counter_offset) & sc_mask32[f.start_counter_length];
								time_data = (uiData[0] >> f.sum_offset) & sc_mask32[f.sum_length];
								diff2 = (uiData[0] >> f.dif2_offset) & sc_mask32[f.dif2_length];
								diff1 = (uiData[0] >> f.dif1_offset) & sc_mask32[f.dif1_length];
								dData = time_data * bs;
								if(output_to_console)
									printf("%d\t%d\t%d\t%d\t%f\t%d\t%d\n", iter_counter, channel, start_counter, time_data, dData, diff2, diff1);
							//	if(output_to_file)
									fprintf(output_file, "%d\t%d\t%d\t%d\t%f\t%d\t%d\n", iter_counter, channel, start_counter, time_data, dData, diff2, diff1);
							}
							else
							{ // tdc mode
								channel = (uiData[0] >> f.channel_offset) & sc_mask32[f.channel_length];
								start_counter = (uiData[0] >> f.start_counter_offset) & sc_mask32[f.start_counter_length];
								sign_counter = (uiData[0] >> f.sign_counter_offset) & sc_mask32[f.sign_counter_length];
								time_data = (uiData[0] >> f.time_data_offset) & sc_mask32[f.time_data_length];
								dData = time_data * bs;
								if(output_to_console)
									printf("%d\t%d\t%d\t%d\t%08x\t%f\n", iter_counter, channel, sign_counter, start_counter, time_data, dData);
							//	if(output_to_file)
									fprintf(output_file, "%d\t%d\t%d\t%d\t%08x\t%f\n", iter_counter, channel, sign_counter, start_counter, time_data, dData);
							}
						}
						if(pipe_empty_returns > 0) pipe_empty_returns = 0;
					}
					else
					{
						progress_bar ++;
						if(pipe_empty_returns >= 0) pipe_empty_returns ++;
					}
				}
				if(progress_bar > 100)
				{
					if(!output_to_console)
					{ // progress stick turning
						turn_progress_bar();
					}
					if(kbhit())
					{ // user has pressed a key
						switch(getch())
						{
						case 'i': // interrupt
							ret = sc_tdc_interrupt2(device_id);
							if(ret < 0)
							{
								printf("interrupt: ");
								eprint(ret);
								return ret;
							}
						case 't': // autorestart off
							auto_restart = false;
							break;
						default : ;
						}
					}
					progress_bar = 0;
				}
			} while((pipe_empty_returns < 0) || (buf_filling > 0));

			if(output_file) fclose(output_file);
		}


		if(subscribe_mask & BUFFER_PIPE)
		{
			for(int pipe_i=0; pipe_i<pipe_count; pipe_i++)
			{ // open files for all pipes
				pipe_buf_output_file[pipe_i] = NULL;
				std::ostringstream convert;
				convert << std::setw(2) << std::setfill('0') << pipe_i;
				switch(pipe_types[pipe_i])
				{
				case TDC_HISTO:
					if(acq_counter)
						pipe_buf_output_file[pipe_i] = fopen((output_file_name + "_tdc_histo" + convert.str() + ".dat").c_str(), "a");
					else
						pipe_buf_output_file[pipe_i] = fopen((output_file_name + "_tdc_histo" + convert.str() + ".dat").c_str(), "w");
					break;
				default: ;
				}
				if(pipe_buf_output_file[pipe_i] == NULL)
				{
					printf("Could not open file\n");
					return -1;
				}
				fprintf(pipe_buf_output_file[pipe_i], "acquisition %d\n", acq_counter);
				accu_counter[pipe_i] = 0;
			}
			unsigned char all_empty;
			do
			{ // repeat reading until the end of the acquisition
				all_empty = 1;
				for(int pipe_i=0; pipe_i<pipe_count; pipe_i++)
				{ // go through all pipes
					if(kbhit())
					{ // user has pressed a key
						switch(getch())
						{
						case 'i': // interrupt
							auto_restart = false;
							ret = sc_tdc_interrupt2(device_id);
							if(ret < 0)
							{
								printf("interrupt: ");
								eprint(ret);
								return ret;
							}
						case 't': // autorestart off
							auto_restart = false;
							break;
						default : ;
						}
					}
					switch(pipe_types[pipe_i])
					{
					case TDC_HISTO:
					{
						histo_size = ((sc_pipe_tdc_histo_params_t *)pipe_params[pipe_i])->size;
						void *buf;
						ret = sc_pipe_read2(device_id, pipe_ids[pipe_i], &buf, ((sc_pipe_tdc_histo_params_t *)pipe_params[pipe_i])->accumulation_ms << 1);
						if(ret < 0)
						{
							fprintf(logfile, "accumulation %d: pipe read ret = %d\n", accu_counter[pipe_i], ret);
							fflush(logfile);
							Beep(660, 50);
						}
						else
						{
							unsigned int i;
							all_empty = 0;
							fprintf(pipe_buf_output_file[pipe_i], "accumulation %d\n", accu_counter[pipe_i]);
							switch(((sc_pipe_tdc_histo_params_t *)pipe_params[pipe_i])->depth)
							{
							case BS8:
								for (i=0; i<histo_size; i++)
								{
									if(((unsigned char *)buf)[i])
										fprintf(pipe_buf_output_file[pipe_i], "%d\t%d\n", i, ((unsigned char *)buf)[i]);
								}
								break;
							case BS16:
								for (i=0; i<histo_size; i++)
								{
									if(((unsigned short *)buf)[i])
										fprintf(pipe_buf_output_file[pipe_i], "%d\t%d\n", i, ((unsigned short *)buf)[i]);
								}
								break;
							case BS32:
								for (i=0; i<histo_size; i++)
								{
									if(((unsigned int *)buf)[i])
										fprintf(pipe_buf_output_file[pipe_i], "%d\t%d\n", i, ((unsigned int *)buf)[i]);
								}
								break;
							case BS64:
								for (i=0; i<histo_size; i++)
								{
									if(((unsigned long long *)buf)[i])
										fprintf(pipe_buf_output_file[pipe_i], "%d\t%d\n", i, ((unsigned long long *)buf)[i]);
								}
								break;
							}
							accu_counter[pipe_i] ++;
						}
						break;
					}
					case STATISTICS:
					{
						void *buf;
						unsigned int i;
						sc_pipe_read2(device_id, pipe_ids[pipe_i], &buf, 1);
					//	output_file = fopen((output_file_name + "_statistics" + convert.str() + ".txt").c_str(), "w");
					//	for (i=0; i<256; i++)
					//		fprintf(output_file, "%08x\n", ((unsigned int *)buf)[i]);
					//	if(output_file) fclose(output_file);
						printf("%d\t%d\t%d\t%d\t\n", ((unsigned int *)buf)[0],((unsigned int *)buf)[4],((unsigned int *)buf)[8],((unsigned int *)buf)[12]);
						printf("%d\t%d\t%d\t%d\t\n", ((unsigned int *)buf)[16],((unsigned int *)buf)[20],((unsigned int *)buf)[24],((unsigned int *)buf)[28]);
						printf("%d\t%d\t%d\t%d\t\n", ((unsigned int *)buf)[128],((unsigned int *)buf)[129],((unsigned int *)buf)[136],((unsigned int *)buf)[137]);
						break;
					}
					}//switch
				}//for
			}while((pipe_empty_returns < 0) || (all_empty == 0));
			for(int pipe_i=0; pipe_i<pipe_count; pipe_i++)
			{ // close files for all pipes
				if(pipe_buf_output_file[pipe_i]) fclose(pipe_buf_output_file[pipe_i]);
				pipe_buf_output_file[pipe_i] = NULL;
			}
		}//if BUFFER_PIPE
		acq_counter ++;
	}


// unsubscribe
	ret = unsubscribe(subscribe_mask);
	if(ret < 0)
	{
		printf("Unubscribing has failed\n");
		return ret;
	}

	ret = sc_tdc_deinit2(device_id);
	if(ret < 0)
	{
		printf("deinit: ");
		eprint(ret);
		return ret;
	}
	fclose(logfile);
	return 0;
}

